from functools import reduce
from pathlib import Path
from typing import List, Tuple

from polylith import check, distributions, libs


def extract_libs_from_project_lock_file(project_data: dict) -> dict:
    lock_file_data = libs.pick_lock_file(project_data["path"])

    if not lock_file_data:
        return {}

    filename = lock_file_data["filename"]
    filetype = lock_file_data["filetype"]

    third_party_libs = libs.extract_libs(project_data, filename, filetype)

    return {"deps": {"items": third_party_libs, "source": filename}}


def extract_libs_from_workspace_lock_file(root: Path, project_data: dict) -> dict:
    lock_file_data = libs.pick_lock_file(root)

    if not lock_file_data:
        return {}

    filename = lock_file_data["filename"]
    filetype = lock_file_data["filetype"]

    data = libs.get_workspace_enabled_lock_file_data(root, filename, filetype)

    if not data:
        return {}

    third_party_libs = libs.extract_workspace_member_libs(data, project_data)

    return {"deps": {"items": third_party_libs, "source": filename}}


def extract_libs_from_lock_file(root: Path, project_data: dict) -> dict:
    from_project = extract_libs_from_project_lock_file(project_data)

    return from_project or extract_libs_from_workspace_lock_file(root, project_data)


def with_third_party_libs_from_lock_file(root: Path, project_data: dict) -> dict:
    third_party_libs = extract_libs_from_lock_file(root, project_data)

    return {**project_data, **third_party_libs}


def check_libs_versions(
    projects_data: List[dict], all_projects_data: List[dict], options: dict
) -> bool:
    is_strict = options["strict"]
    is_quiet = options["quiet"]

    if not is_strict:
        return True

    development_data = next(p for p in all_projects_data if p["type"] == "development")

    libraries = libs.report.libs_with_different_versions(
        development_data, projects_data
    )

    if not is_quiet:
        libs.report.print_libs_with_different_versions(
            libraries, development_data, projects_data, options
        )

    return False if libraries else True


def run_each(
    root: Path, ns: str, project_data: dict, options: dict
) -> Tuple[bool, dict]:
    is_quiet = options["quiet"]
    is_strict = options["strict"]

    name = project_data["name"]
    deps = project_data["deps"]
    alias = options["alias"]

    from_lock_file = libs.is_from_lock_file(deps)

    collected_imports = check.report.collect_all_imports(root, ns, project_data)
    collected_libs = distributions.known_aliases_and_sub_dependencies(
        deps,
        alias,
        options,
        from_lock_file,
    )

    details = check.report.create_report(
        project_data,
        collected_imports,
        collected_libs,
        is_strict or from_lock_file,
    )

    res = all([not details["brick_diff"], not details["libs_diff"]])

    if not is_quiet:
        check.report.print_missing_deps(details["brick_diff"], name)
        check.report.print_missing_deps(details["libs_diff"], name)

    return res, details


def _merge(data: List[dict], key: str) -> dict:
    return reduce(lambda acc, d: {**acc, **d[key]}, data, {})


def _print_brick_imports(all_imports: List[dict]) -> None:
    merged_bases = _merge(all_imports, "bases")
    merged_components = _merge(all_imports, "components")

    merged = {"bases": merged_bases, "components": merged_components}

    check.report.print_brick_imports(merged)


def run(root: Path, ns: str, projects_data: List[dict], options: dict) -> bool:
    is_verbose = options["verbose"] and not options["quiet"]
    res = [run_each(root, ns, p, options) for p in projects_data]
    results = [r[0] for r in res]

    if is_verbose:
        brick_imports = [r[1]["brick_imports"] for r in res]
        third_party_imports = [r[1]["third_party_imports"] for r in res]

        _print_brick_imports(brick_imports)
        _print_brick_imports(third_party_imports)

    return all(results)



from polylith.project import templates
from polylith.project.create import create_project
from polylith.project.get import (
    get_packages_for_projects,
    get_project_name,
    get_project_template,
    get_toml,
)
from polylith.project.parser import parse_package_paths

__all__ = [
    "create_project",
    "get_packages_for_projects",
    "get_project_name",
    "get_project_template",
    "get_toml",
    "parse_package_paths",
    "templates",
]



from functools import reduce
from pathlib import Path
from typing import List

from polylith.toml import load_toml

patterns = {
    "pdm.lock": "toml",
    "poetry.lock": "toml",
    "uv.lock": "toml",
    "requirements.lock": "text",
    "requirements.txt": "text",
}


def find_lock_files(path: Path) -> dict:
    return {k: v for k, v in patterns.items() if Path(path / k).exists()}


def pick_lock_file(path: Path) -> dict:
    data = find_lock_files(path)
    first = next(iter(data.items()), None)

    if not first:
        return {}

    filename, filetype = first

    return {"filename": filename, "filetype": filetype}


def extract_libs_from_packages(packages: List[dict]) -> dict:
    return {p["name"]: p["version"] for p in packages}


def extract_libs_from_toml(path: Path) -> dict:
    data = load_toml(path)

    return extract_libs_from_packages(data.get("package", []))


def parse_name(row: str) -> str:
    parts = str.split(row, "==")

    return parts[0]


def parse_version(row: str) -> str:
    parts = str.split(row, "==")[1]
    res = str.split(parts, " ")

    return res[0]


def extract_libs_from_txt(path: Path) -> dict:
    with open(path, "r") as f:
        data = f.readlines()

    rows = (str.strip(line) for line in data)
    filtered = (row for row in rows if row and not row.startswith(("#", "-")))

    return {parse_name(row): parse_version(row) for row in filtered}


def extract_libs(project_data: dict, filename: str, filetype: str) -> dict:
    dir_path = project_data["path"]

    path = Path(dir_path / filename)

    if not path.exists():
        return {}

    try:
        if filetype == "toml":
            return extract_libs_from_toml(path)

        return extract_libs_from_txt(path)
    except (IndexError, KeyError, ValueError) as e:
        raise ValueError(f"Failed reading {filename}: {repr(e)}") from e


def is_from_lock_file(deps: dict) -> bool:
    return any(deps["source"] == s for s in set(patterns.keys()))


def get_workspace_members(data: dict) -> List[str]:
    return data.get("manifest", {}).get("members", [])


def get_workspace_enabled_lock_file_data(
    root: Path, filename: str, filetype: str
) -> dict:
    if filetype != "toml":
        return {}

    path = Path(root / filename)

    if not path.exists():
        return {}

    data = load_toml(path)
    members = get_workspace_members(data)

    return data if members else {}


def pick_packages(data: dict, name: str) -> list:
    package = next(p for p in data["package"] if p["name"] == name)

    package_sub_deps = package.get("dependencies", [])
    nested_package_deps = [pick_packages(data, p["name"]) for p in package_sub_deps]

    flattened = sum(nested_package_deps, [])

    return [package] + flattened if flattened else [package]


def normalized(name: str) -> str:
    chars = {"_", "."}

    normalized = reduce(lambda acc, char: str.replace(acc, char, "-"), chars, name)

    return str.lower(normalized)


def extract_workspace_member_libs(data: dict, project_data: dict) -> dict:
    members = get_workspace_members(data)
    member_name = normalized(project_data["name"])

    if member_name not in members:
        return {}

    try:
        packages = pick_packages(data, member_name)
        extracted = extract_libs_from_packages(packages)
    except KeyError as e:
        raise ValueError(f"Failed parsing lock-file data: {repr(e)}") from e

    return {k: v for k, v in extracted.items() if k != member_name}



from polylith.libs import report
from polylith.libs.grouping import extract_third_party_imports, get_third_party_imports
from polylith.libs.lock_files import (
    extract_libs,
    extract_workspace_member_libs,
    get_workspace_enabled_lock_file_data,
    is_from_lock_file,
    pick_lock_file,
)

__all__ = [
    "report",
    "extract_third_party_imports",
    "get_third_party_imports",
    "extract_libs",
    "extract_workspace_member_libs",
    "get_workspace_enabled_lock_file_data",
    "is_from_lock_file",
    "pick_lock_file",
]
